# OALABS
#
# Much of this code was copied from or heavily influnced by Mario Vilas 
# winappdebug project. I am including the full winappdebug copyright notice
# and a link to the project on GitHub.
#
# https://github.com/MarioVilas/winappdbg
#
#
#
# Copyright (c) 2009-2020, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import ctypes
# Imports from pywin32
import win32api
import win32security
# Local imports
from . import win32types
from . import win32utils


# Copy paste from winappdbg: https://github.com/MarioVilas/winappdbg
# BOOL WINAPI ReadProcessMemory(
#   __in   HANDLE hProcess,
#   __in   LPCVOID lpBaseAddress,
#   __out  LPVOID lpBuffer,
#   __in   SIZE_T nSize,
#   __out  SIZE_T* lpNumberOfBytesRead
# );
# + Maybe change page permissions before trying to read?
# Thomas: Page permissions might be a problem only for edge cases, like MEM_NOACCESS
def read(hProcess, lpBaseAddress, nSize):
    _ReadProcessMemory = ctypes.windll.kernel32.ReadProcessMemory
    _ReadProcessMemory.argtypes = [win32types.HANDLE, 
                                    win32types.LPVOID, 
                                    win32types.LPVOID, 
                                    win32types.SIZE_T, 
                                    win32types.POINTER(win32types.SIZE_T)]
    _ReadProcessMemory.restype  = bool

    lpBuffer            = ctypes.create_string_buffer(b'', nSize)
    lpNumberOfBytesRead = win32types.SIZE_T(0)
    success = _ReadProcessMemory(hProcess, lpBaseAddress, lpBuffer, nSize, ctypes.byref(lpNumberOfBytesRead))
    if not success and ctypes.WinError().errno != win32types.ERROR_PARTIAL_COPY:
        raise ctypes.WinError()
    return (lpBuffer.raw)[:lpNumberOfBytesRead.value]


# BOOL WINAPI WriteProcessMemory(
#   __in   HANDLE hProcess,
#   __in   LPCVOID lpBaseAddress,
#   __in   LPVOID lpBuffer,
#   __in   SIZE_T nSize,
#   __out  SIZE_T* lpNumberOfBytesWritten
# );
def WriteProcessMemory(hProcess, lpBaseAddress, lpBuffer):
    _WriteProcessMemory = ctypes.windll.kernel32.WriteProcessMemory
    _WriteProcessMemory.argtypes = [win32types.HANDLE, 
                                    win32types.LPVOID, 
                                    win32types.LPVOID, 
                                    win32types.SIZE_T, 
                                    win32types.POINTER(win32types.SIZE_T)]
    _WriteProcessMemory.restype  = bool

    nSize                   = len(lpBuffer)
    lpBuffer                = ctypes.create_string_buffer(lpBuffer)
    lpNumberOfBytesWritten  = win32types.SIZE_T(0)
    success = _WriteProcessMemory(hProcess, lpBaseAddress, lpBuffer, nSize, ctypes.byref(lpNumberOfBytesWritten))
    if not success and ctypes.WinError().errno != win32types.ERROR_PARTIAL_COPY:
        raise ctypes.WinError()
    return lpNumberOfBytesWritten.value


# SIZE_T WINAPI VirtualQueryEx(
#   __in      HANDLE hProcess,
#   __in_opt  LPCVOID lpAddress,
#   __out     PMEMORY_BASIC_INFORMATION lpBuffer,
#   __in      SIZE_T dwLength
# );
def VirtualQueryEx(hProcess, lpAddress):
    _VirtualQueryEx = ctypes.windll.kernel32.VirtualQueryEx
    _VirtualQueryEx.argtypes = [win32types.HANDLE, 
                                win32types.LPVOID, 
                                win32types.POINTER(win32types.MEMORY_BASIC_INFORMATION), 
                                win32types.SIZE_T]
    _VirtualQueryEx.restype  = win32types.SIZE_T

    lpBuffer  = win32types.MEMORY_BASIC_INFORMATION()
    dwLength  = ctypes.sizeof(win32types.MEMORY_BASIC_INFORMATION)
    success   = _VirtualQueryEx(hProcess, lpAddress, ctypes.byref(lpBuffer), dwLength)
    if success == 0:
        raise ctypes.WinError()
    return lpBuffer


# BOOL WINAPI VirtualProtectEx(
#   __in   HANDLE hProcess,
#   __in   LPVOID lpAddress,
#   __in   SIZE_T dwSize,
#   __in   DWORD flNewProtect,
#   __out  PDWORD lpflOldProtect
# );
def VirtualProtectEx(hProcess, lpAddress, dwSize, flNewProtect = win32types.PAGE_EXECUTE_READWRITE):
    _VirtualProtectEx = ctypes.windll.kernel32.VirtualProtectEx
    _VirtualProtectEx.argtypes = [  win32types.HANDLE, 
                                    win32types.LPVOID, 
                                    win32types.SIZE_T, 
                                    win32types.DWORD, 
                                    win32types.PDWORD]
    _VirtualProtectEx.restype  = bool
    _VirtualProtectEx.errcheck = win32utils.RaiseIfZero

    flOldProtect = win32types.DWORD(0)
    _VirtualProtectEx(hProcess, lpAddress, dwSize, flNewProtect, ctypes.byref(flOldProtect))
    return flOldProtect.value



def write(hProcess, lpAddress, lpBuffer):
    # TODO: this should work without the mem change wrapper
    #       but it doesn'!! WHY??
    #
    #       https://devblogs.microsoft.com/oldnewthing/20181206-00/?p=100415
    #
    bytes_written = 0
    pMemoryBasicInformation = VirtualQueryEx(hProcess, lpAddress)
    if (win32types.WRITEABLE & pMemoryBasicInformation.Protect) != 0:
        # It is writable let's write
        bytes_written = WriteProcessMemory(hProcess, lpAddress, lpBuffer)
    else:
        # Not writable let's fix this temporarily
        old_protections = VirtualProtectEx(hProcess, lpAddress, len(lpBuffer), win32types.PAGE_READWRITE)
        bytes_written = WriteProcessMemory(hProcess, lpAddress, lpBuffer)
        new_protections = VirtualProtectEx(hProcess, lpAddress, len(lpBuffer), old_protections)
    return bytes_written




